/***************************************************************************/
/*  Copyright 1994 National Instruments Corporation.  All Rights Reserved  */
/***************************************************************************/

#include <utility.h>
#include <gpib.h>
#include <formatio.h>
#include "ke2001.h"

/* = Keithley 2001 Scanning Digital Multi-Meter ============================ */
/*  LabWindows 2.0 Instrument Driver                                         */
/*  Original Release: Feb, 1993                                              */
/*  By:    HG & GD, LWILDP,                                                  */
/*         National Instruments, Austin, TX. (800) 433-3488                  */
/*  Originally written in C                                                  */
/*  Modification History: The selections for function type in the Averaging  */
/*                        Filter function were reversed.  The setup arrays   */
/*                        have been modified, so that 0 corresponds to       */
/*                        repeat and 1 corresponds to moving.                */
/*                                                                           */
/*  Modification Date:    December 13, 1993                                  */
/*                                                                           */
/*  Modified by:          GD                                                 */
/*                        National Instruments, Austin, Texas                */
/*                        Phone (800) 433-3488  Fax (512) 794-5678           */
/* ========================================================================= */
/*                                                                           */
/*                        Changed the GPIB Library calls so that they check  */
/*                        for the errorbit being set to indicate an error    */
/*                        (& 0x8000)instead of checking for a negative       */
/*                        return value.                                      */
/*                                                                           */
/*  Modification Date:    April 20, 1994                                     */
/*                                                                           */
/*  Modified by:          JM                                                 */
/*                        National Instruments, Austin, Texas                */
/*                        Phone (800) 433-3488  Fax (512) 794-5678           */
/*========================================================================== */
/*                                                                           */
/*                        Added codes and changed fuction prototypes to      */
/*                        create DLL                                         */
/*                                                                           */
/*  Modification Date:    October 22, 1994                                   */
/*  Modified by:          SA,  LWILDP                                        */
/*                                                                           */
/*  June 1996:            Modified for MDLL compatibility.                   */
/*                        Modified by:  TKN, National Instruments            */
/*                        Phone (800) 433-3488  Fax (512) 794-5678           */    
/*                                                                           */
/* ========================================================================= */

/* = DEFINE STATEMENTS ===================================================== */
#define MAX_CMD_LENGTH  1000

/* = UTILITY ROUTINES ====================================================== */
int kei2001_open_instr (int, int *);
int kei2001_close_instr (int);
int kei2001_invalid_real_range (double, double, double, int);
int kei2001_invalid_integer_range (int, int, int, int);
int kei2001_invalid_long_range (long, long, long, int);
int kei2001_device_closed (int);
int kei2001_read_data (int, char *, int);
int kei2001_write_data (int, char *, int);
void kei2001_setup_arrays (void);
int kei2001_check_instr_status (int);
int kei2001_read_data_file (int, char *);
int kei2001_loc_remote (int, int);

/* = INSTRUMENT TABLE ====================================================== */
/*  address array: contains the GPIB addresses of opened instruments.        */
/*  bd array: contains the device descriptors returned by OpenDev.           */
/*  instr_cnt: contains the number of instruments open of this model type.   */
/*  kei2001_err: the error variable for the instrument module                 */
/* ========================================================================= */
/*   cmd is a buffer for GPIB I/O strings.                                     */
/*  =========================================================================  */
static char cmd[1000];
static int kei2001_scanner_present[kei2001_MAX_INSTR + 1];
static int kei2001_memory_size[kei2001_MAX_INSTR + 1];
static int address[kei2001_MAX_INSTR + 1];
static int bd[kei2001_MAX_INSTR + 1];
static int instr_cnt;
static int kei2001_err;

/* = INSTRUMENT-DEPENDENT COMMAND ARRAYS =================================== */
static char *function[8];
static double func_input_range[6];
static char *scan_func[6];
static char *trig_cmd[8];
static char *dc_str[2];
static char *freq_source_str[2];
static char *tc_type_str;
static char *filter_type_str[2];
static char *rtd_type_str[2];
static char *trans_type_str[3];
static char *temp_unit_str;
static char *ac_unit_str[3];
static char *auto_zero_mode_str[2];
static char *auto_filter_str[3];
static char *acq_method_str[2];
static char *source_str[7];
static char *egroup_str[2];
static char *buf_src_str[3];
static char *pretrig_src_str[4];
static char *buf_cntrl_mode_str[4];
static char *math_format_str[3];
static char *calc2_format_str[6];
static char *chan_op_str[4];
static char *chan_fun_str[7];
static char *source_bypass_str[2];
static char *rat_del_fun_str[3];
static char *polarity_str[2];
static double freq_def_range[2];
static double dbm_imp[6];
static double freq_source_range[2];
static double aperture[4];
static long buf_size_range[6];
static int sav_size[3];

/*  =========================================================================  */
/*   This function opens the instrument, queries the instrument for its ID,    */
/*   and initializes the instrument to a known state.                          */
/*  =========================================================================  */
int PUBLIC kei2001_init (int addr,int resetDev,int *instrID)
{
    int id;
    char mem[10];
    char scanner[20];
    char header[7];

    if (kei2001_invalid_integer_range (addr, 0, 30,  -1) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (resetDev, 0, 1,  -2) != 0)
        return kei2001_err;
    /*  Initialize entry in Instrument Table and interface for instrument.  */
    if (kei2001_open_instr (addr, &id) != 0)
        return kei2001_err;
    /*  Perform an Identification Query on the instrument.  */
    if (kei2001_write_data (id, "*IDN?", 5) != 0)  {
        kei2001_close_instr (id);
        return kei2001_err;
    }
    if (kei2001_read_data (id, cmd, MAX_CMD_LENGTH) != 0)  {
        kei2001_close_instr (id);
        return kei2001_err;
    }
    /*  Check to see if the ID query returns the correct response.  */
    Scan (cmd, "%s>KEITHLEY INSTRUMENTS INC.,MODEL 2001");
    if (NumFmtdBytes () != 36)  {
        kei2001_err = 223;
        kei2001_close_instr (id);
        return kei2001_err;
    }
    if (kei2001_write_data (id, "*OPT?", 5) != 0)  {
        kei2001_close_instr (id);
        return kei2001_err;
    }
    FillBytes (cmd, 0, 100, 0);
    if (kei2001_read_data (id, cmd, 17) != 0)  {
        kei2001_close_instr (id);
        return kei2001_err;
    }
    Scan (cmd, "%s>%s[t44],%s", mem, scanner);
    Scan (scanner, "%s>2001-SCAN");
    if (NumFmtdBytes () == 9)
        kei2001_scanner_present[id] = 1;
    else
        kei2001_scanner_present[id] = 0;
    Scan (mem, "%s>%s[t#]%d[b4]", header, &kei2001_memory_size[id]);
    kei2001_setup_arrays ();
    if (resetDev != 0)  {
        if (kei2001_write_data (id, "*RST\r\n", 6) != 0)
            return kei2001_err;
        Fmt (cmd, "%s<*CLS;*ESE 1;*OPC;");
        if (kei2001_write_data (id, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    /*  Check the Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (id) != 0)
        return kei2001_err;
    *instrID = id;
    return kei2001_err;
}

/*  =========================================================================  */
/*   This changes the current instrument function.                             */
/*  =========================================================================  */
int PUBLIC kei2001_change_func (int instrID,int fun)
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (fun, 0, 7,  -2) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:func \'%s\'", function[fun]);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    /*  Check the Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*  =========================================================================  */
/*   This sets a scan channel for the selected measurement function                 */
/*  =========================================================================  */
int PUBLIC kei2001_conf_chan (int instrID,int chan,int fun)
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (chan, 1, 10,  -2) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (fun, 0, 3,  -3) != 0 && fun != 6 && fun != 7)  {
        kei2001_err =  -3;
        return kei2001_err;
    }
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:rout:scan:func (@%d[b4]), \'%s\'", chan, function[fun]);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    /*  Check the Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*  =========================================================================  */
/*   This sets the configuration for the selected measurement function         */
/*  =========================================================================  */
int PUBLIC kei2001_conf_volt_curr_res (int instrID,int fun,double range,
                                    int res,int dc,double int_time)
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (fun, 0, 5,  -2) != 0)
        return kei2001_err;
    if (range != -1.0)
        if (kei2001_invalid_real_range (range, 0.0, func_input_range[fun],  -3) != 0)
            return kei2001_err;
    if (res != 0)
        if (kei2001_invalid_integer_range (res, 4, 8,  -4) != 0)
            return kei2001_err;
    if (kei2001_invalid_integer_range (dc, 0, 1,  -5) != 0)
        return kei2001_err;
    if (int_time != -1.0)
        if (kei2001_invalid_real_range (int_time, 0.01, 10.0,  -6) != 0)
            return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    if (range == -1.0)  {
        Fmt (cmd, "%s<:%s:rang:auto 1\r\n", function[fun]);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    else  {
        Fmt (cmd, "%s<:%s:rang %f\r\n", function[fun], range);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    if (!res)  {
        Fmt (cmd, "%s<:%s:dig:auto 1\r\n", function[fun]);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    else  {
        Fmt (cmd, "%s<:%s:dig %d[b4]\r\n", function[fun], res);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    if (int_time == -1.0)  {
        Fmt (cmd, "%s<:%s:NPLC:auto 1\r\n", function[fun]);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    else  {
        Fmt (cmd, "%s<:%s:NPLC %f\r\n", function[fun], int_time);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    if (fun == 1 || fun == 5)  {
        Fmt (cmd, "%s<:%s:coup %s\r\n", function[fun], dc_str[dc]);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    /*  Check the Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*  =========================================================================  */
/*   This sets the configuration for the frequency measurement function        */
/*  =========================================================================  */
int PUBLIC kei2001_conf_freq (int instrID,int res,int coup,int src, double range, 
                            double level,int ttl)
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (res, 4, 5,  -2) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (coup, 0, 1,  -3) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (src, 0, 1,  -4) != 0)
        return kei2001_err;
    if (range != -1.0)  {
        if (kei2001_invalid_real_range (range, 0.0, freq_source_range[src],  -5) != 0)
            return kei2001_err;
        if (kei2001_invalid_real_range (level, -0.6 * range, 0.6 * range,  -6) != 0)
            return kei2001_err;
    }
    else  if (kei2001_invalid_real_range (level, -0.6 * freq_def_range[src], 0.6 * freq_def_range[src],  -6) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (ttl, 0, 1,  -7) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    if (!res)  {
        Fmt (cmd, "%s<:freq:dig:auto 1\r\n");
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    else  {
        Fmt (cmd, "%s<:freq:dig %d[b4]\r\n", res);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    Fmt (cmd, "%s<:freq:coup %s\r\n", dc_str[coup]);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:freq:sour %s\r\n", freq_source_str[src]);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    if (range == -1.0)  {
        Fmt (cmd, "%s<:freq:thr:%s:rang %f\r\n", freq_source_str[src], freq_def_range[src]);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    else  {
        Fmt (cmd, "%s<:freq:thr:%s:rang %f\r\n", freq_source_str[src], range);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    if (!src && ttl != 0)  {
        Fmt (cmd, "%s<:freq:thr:volt:ttl\r\n");
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    /*  Check the Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*  =========================================================================  */
/*   This sets the configuration for the temperature measurement function      */
/*  =========================================================================  */
int PUBLIC kei2001_conf_temp (int instrID,double int_time,int res,int trans, 
                            int rtd,int tc)
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (int_time != -1.0)
        if (kei2001_invalid_real_range (int_time, 0.01, 10.0,  -2) != 0)
            return kei2001_err;
    if (res != 0)
        if (kei2001_invalid_integer_range (res, 4, 7,  -3) != 0)
            return kei2001_err;
    if (kei2001_invalid_integer_range (trans, 0, 2,  -4) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (rtd, 0, 1,  -5) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (tc, 0, 6,  -6) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    if (int_time == -1.0)  {
        Fmt (cmd, "%s<:temp:nplc:auto 1\r\n");
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    else  {
        Fmt (cmd, "%s<:temp:nplc %f\r\n", int_time);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    if (!res)  {
        Fmt (cmd, "%s<:temp:dig:auto 1\r\n");
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    else  {
        Fmt (cmd, "%s<:temp:dig %d[b4]\r\n", res);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    Fmt (cmd, "%s<:temp:tran %s\r\n", trans_type_str[trans]);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    if (!trans || trans == 1)  {
        Fmt (cmd, "%s<:temp:rtd:type %s\r\n", rtd_type_str[rtd]);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    else  {
        Fmt (cmd, "%s<:temp:tc:type %c\r\n", tc_type_str[tc]);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    /*  Check the Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*  ===========================================================================  */
/*   This function configures units for ac voltage and temperature measurements  */
/* ============================================================================  */
int PUBLIC kei2001_conf_unit (int instrID,int temp,int ac_volt,double db_ref, 
                            int dbm)
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (temp, 0, 2,  -2) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (ac_volt, 0, 2,  -3) != 0)
        return kei2001_err;
    if (kei2001_invalid_real_range (db_ref, 0.01, 750.0,  -4) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (dbm, 0, 6,  -5) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:unit:temp %c\r\n", temp_unit_str[temp]);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:unit:volt:ac %s\r\n", ac_unit_str[ac_volt]);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    if (ac_volt == 1)  {
        Fmt (cmd, "%s<:unit:volt:ac:db:ref %f\r\n", db_ref);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    else  if (ac_volt == 2)  {
        Fmt (cmd, "%s<:unit:volt:ac:dbm:imp %f\r\n", dbm_imp[dbm]);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    /*  Check the Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*  =========================================================================  */
/*   This function configures the averaging filter                             */
/* ========================================================================== */
int PUBLIC kei2001_conf_ave_filter (int instrID,int fun,int state,int auto_filter, 
                                int type,int count,int adv_filter,double toler)
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (fun, 0, 5,  -2) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (state, 0, 1,  -3) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (auto_filter, 0, 2,  -4) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (type, 0, 1,  -5) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (count, 1, 100,  -6) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (adv_filter, 0, 1,  -7) != 0)
        return kei2001_err;
    if (toler != -1.0)
        if (kei2001_invalid_real_range (toler, 0.0, 100.0,  -8) != 0)
            return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:%s:aver:tcon %s\r\n", function[fun], filter_type_str[type]);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:%s:aver:coun %d[b4]\r\n", function[fun], count);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:%s:aver:adv %d[b4]\r\n", function[fun], adv_filter);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    if (adv_filter != 0)
        if (toler != -1.0)  {
            Fmt (cmd, "%s<:%s:aver:adv:ntol %f\r\n", function[fun], toler);
            if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
                return kei2001_err;
        }
        else  {
            Fmt (cmd, "%s<:%s:aver:adv:ntol def\r\n", function[fun]);
            if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
                return kei2001_err;
        }
    Fmt (cmd, "%s<:%s:aver:auto %s\r\n", function[fun], auto_filter_str[auto_filter]);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:%s:aver %d[b4]\r\n", function[fun], state);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    /*  Check the Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*  =========================================================================  */
/*   This specifies the scanning list for the instrument                       */
/*  =========================================================================  */
int PUBLIC kei2001_conf_scanlist (int instrID,int num_chan,int chan[])
{
    int i;
    char temp[50];

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (num_chan, 1, 10,  -2) != 0)
        return kei2001_err;
    for (i = 0; i < num_chan; i++)
        if (kei2001_invalid_integer_range (chan[i], 1, 10,  -3) != 0)
            return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    Fmt (temp, "%s<:rout:scan (@%d[b4]", chan[0]);
    for (i = 1; i < num_chan; i++)
        Fmt (temp, "%s<%s,%d[b4]", temp, chan[i]);
    Fmt (cmd, "%s<%s)\r\n", temp);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    /*  Check the Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*  =========================================================================  */
/*   This function configures the ratio/delta operation between two channels   */
/* ==========================================================================  */
int PUBLIC kei2001_conf_rat_del (int instrID,int op,int fun,int meas_chan, int ref_chan)
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (op, 0, 1,  -2) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (fun, 0, 1,  -3) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (ref_chan, 1, 10,  -4) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (meas_chan, 1, 10,  -5) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    if (!op)  {
        Fmt (cmd, "%s<:rout:scan:rat:func \'%s\'\r\n", rat_del_fun_str[fun]);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
        Fmt (cmd, "%s<:rout:scan:rat:rch (@ %d[b4])\r\n", ref_chan);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
        Fmt (cmd, "%s<:rout:scan:rat:mch (@ %d[b4])\r\n", meas_chan);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    else  if (op == 1)  {
        Fmt (cmd, "%s<:rout:scan:delt:func \'%s\'\r\n", rat_del_fun_str[fun]);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
        Fmt (cmd, "%s<:rout:scan:delt:rch (@ %d[b4])\r\n", ref_chan);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
        Fmt (cmd, "%s<:rout:scan:delt:mch (@ %d[b4])\r\n", meas_chan);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    /*  Check the Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/* ========================================================================== */
/*  Function: Configure Math                                                  */
/*  Description: Specify the configuration of the math operation              */
/* ========================================================================== */
int PUBLIC kei2001_conf_math (int instrID,int state,int frmt,double pf,double mf,double bf)
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (state, 0, 1,  -2) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (frmt, 0, 2,  -3) != 0)
        return kei2001_err;
    if (kei2001_invalid_real_range (pf, -2000000000.0, 2000000000.0,  -4) != 0)
        return kei2001_err;
    if (kei2001_invalid_real_range (mf, -9.999999e35, 9.999999e35,  -5) != 0)
        return kei2001_err;
    if (kei2001_invalid_real_range (bf, -9.999999e35, 9.999999e35,  -6) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:calc:stat %d[b4]\r\n", state);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:calc:form %s\r\n", math_format_str[frmt]);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    if (!frmt)  {
        Fmt (cmd, "%s<:calc:kmat:mmf %f\r\n", mf);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
        Fmt (cmd, "%s<:calc:kmat:mbf %f\r\n", bf);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    else  if (frmt == 1)  {
        Fmt (cmd, "%s<:calc:kmat:perc %f\r\n", pf);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    /*  Check the Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*  ========================================================================  */
/*   This function configures the auto zero and line synchronization          */
/* =========================================================================  */
int PUBLIC kei2001_conf_syst_setup (int instrID,int az_state,int az_mode,int line_sync)
{
    int cont;

    /*   remember continuous initiation state  */
    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (az_state, 0, 1,  -2) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (az_mode, 0, 1,  -3) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (line_sync, 0, 1,  -4) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    /*  first put the instrument into the idle state  */
    Fmt (cmd, "%s<:init:cont?\r\n");
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    if (kei2001_read_data (instrID, cmd, 100) != 0)
        return kei2001_err;
    Scan (cmd, "%s>%d[b4]", &cont);
    if (cont != 0)  {
        Fmt (cmd, "%s<:init:cont 0\r\n");
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    Fmt (cmd, "%s<:abor\r\n");
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:syst:azer:stat %d[b4]\r\n", az_state);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:syst:azer:type %s\r\n", auto_zero_mode_str[az_mode]);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:syst:lsyn:stat %d[b4]\r\n", line_sync);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    if (cont != 0)  {
        Fmt (cmd, "%s<:init:cont 1\r\n");
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    /*  Check the Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*  ========================================================================  */
/*   This function configures the reading acquisition method of the           */
/*   instrument                                                               */
/* =========================================================================  */
int PUBLIC kei2001_conf_acq_method (int instrID,int acq_meth)
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (acq_meth, 0, 1,  -1) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:syst:amet %s\r\n", acq_method_str[acq_meth]);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    /*  Check the Event Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*  ========================================================================  */
/*   This function configures the arm layer of the triggering system          */
/* ========================================================================== */
int PUBLIC kei2001_conf_trig_arm (int instrID,int source,int inf, 
                                long count,int oline,int iline,int bypass)
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (source, 0, 5,  -2) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (inf, 0, 1,  -3) != 0)
        return kei2001_err;
    if (kei2001_invalid_long_range (count, 1L, 99999L,  -4) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (oline, 1, 6,  -5) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (iline, 1, 6,  -6) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (bypass, 0, 1,  -7) != 0)
        return kei2001_err;
    if (iline == oline)  {
        kei2001_err = 20;
        /*  input constraint  */
        return kei2001_err;
    }
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:arm:sour %s\r\n", source_str[source]);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    if (inf == 1)  {
        Fmt (cmd, "%s<:arm:coun inf\r\n");
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    else  {
        Fmt (cmd, "%s<:arm:coun %i[b4]\r\n", count);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    Fmt (cmd, "%s<:arm:tcon:asyn:ilin %d[b4]\r\n", iline);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:arm:tcon:asyn:olin %d[b4]\r\n", oline);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:arm:tcon:dir %s\r\n", source_bypass_str[bypass]);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    /*  Check the Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*  ========================================================================  */
/*   This function configures the scan layer of the triggering system         */
/* ========================================================================== */
int PUBLIC kei2001_conf_trig_scan (int instrID,int source,int inf,long count, 
                                int oline,int iline,int bypass,double dela,double tmer)
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (source, 0, 6,  -2) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (inf, 0, 1,  -3) != 0)
        return kei2001_err;
    if (kei2001_invalid_long_range (count, 1L, 99999L,  -4) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (oline, 1, 6,  -5) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (iline, 1, 6,  -6) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (bypass, 0, 1,  -7) != 0)
        return kei2001_err;
    if (kei2001_invalid_real_range (dela, 0.0, 999999.999,  -8) != 0)
        return kei2001_err;
    if (kei2001_invalid_real_range (tmer, 0.001, 999999.999,  -9) != 0)
        return kei2001_err;
    if (iline == oline)  {
        kei2001_err = 20;
        /*  input constraint  */
        return kei2001_err;
    }
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:arm:lay2:sour %s\r\n", source_str[source]);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    if (inf == 1)  {
        Fmt (cmd, "%s<:arm:lay2:coun inf\r\n");
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    else  {
        Fmt (cmd, "%s<:arm:lay2:coun %i[b4]\r\n", count);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    Fmt (cmd, "%s<:arm:lay2:tcon:asyn:ilin %d[b4]\r\n", iline);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:arm:lay2:tcon:asyn:olin %d[b4]\r\n", oline);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:arm:lay2:tcon:dir %s\r\n", source_bypass_str[bypass]);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:arm:lay2:del %f\r\n", dela);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:arm:lay2:tim %f\r\n", tmer);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    /*  Check the Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*  ========================================================================  */
/*   This function configures the measure layer of the triggering system                       */
/* ========================================================================== */
int PUBLIC kei2001_conf_trig_meas (int instrID,int source,int inf,long count,
                                int oline,int iline,double dela,double tmer)
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (source, 0, 6,  -2) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (inf, 0, 1,  -3) != 0)
        return kei2001_err;
    if (kei2001_invalid_long_range (count, 1L, 99999L,  -4) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (oline, 1, 6,  -5) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (iline, 1, 6,  -6) != 0)
        return kei2001_err;
    if (kei2001_invalid_real_range (dela, 0.0, 999999.999,  -7) != 0)
        return kei2001_err;
    if (kei2001_invalid_real_range (tmer, 0.001, 999999.999,  -8) != 0)
        return kei2001_err;
    if (iline == oline)  {
        kei2001_err = 20;
        /*  input constraint  */
        return kei2001_err;
    }
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:trig:sour %s\r\n", source_str[source]);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    if (inf == 1)  {
        Fmt (cmd, "%s<:trig:coun inf\r\n");
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    else  {
        Fmt (cmd, "%s<:trig:coun %i[b4]\r\n", count);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    Fmt (cmd, "%s<:trig:tcon:asyn:ilin %d[b4]\r\n", iline);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:trig:tcon:asyn:olin %d[b4]\r\n", oline);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:trig:del %f\r\n", dela);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:trig:tim %f\r\n", tmer);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    /*  Check the Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*  ========================================================================  */
/*   This function selects the initiation continuity                          */
/* ========================================================================== */
int PUBLIC kei2001_cont_init (int instrID,int cont)
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (cont, 0, 1,  -2) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:init:cont %d[b4]\r\n", cont);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    /*  Check the Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*  ========================================================================  */
/*   This function configures the buffer settings                                      */
/* ========================================================================== */
int PUBLIC kei2001_conf_buf (int instrID,int e_group,int src,int a_size,
                            long b_size,int b_mode,double pre_reading,int pre_src)
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (e_group, 0, 1,  -2) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (src, 0, 2,  -3) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (a_size, 0, 1,  -4) != 0)
        return kei2001_err;
    if (kei2001_invalid_long_range (b_size, 2L, buf_size_range[kei2001_memory_size[instrID] + 3 * e_group],  -5) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (b_mode, 0, 3,  -6) != 0)
        return kei2001_err;
    if (kei2001_invalid_real_range (pre_reading, 0.0, 100.0,  -7) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (pre_src, 0, 3,  -8) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:trac:egr %s\r\n", egroup_str[e_group]);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:trac:feed %s\r\n", buf_src_str[src]);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:trac:poin:auto %d[b4]\r\n", a_size);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    if (!a_size)  {
        Fmt (cmd, "%s<:trac:poin  %i[b4]\r\n", b_size);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    Fmt (cmd, "%s<:trac:feed:cont %s\r\n", buf_cntrl_mode_str[b_mode]);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    if (b_mode == 3)  {
        /*  pretrigger buffer control */
        Fmt (cmd, "%s<:trac:feed:pret:amo:perc %f\r\n", pre_reading);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
        Fmt (cmd, "%s<:trac:feed:pret:sour %s\r\n", pretrig_src_str[pre_src]);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    /*  Check the Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*  ========================================================================  */
/*   This function takes a single reading from the instrument                 */
/* =========================================================================  */
int PUBLIC kei2001_sing_meas (int instrID,int op,int chan,double *result)
{
    int cont;
/*    int adv_filter_dc;     */
/*    int adv_filter_res;    */

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (op, 0, 2,  -2) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (chan, 0, 10,  -3) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    if (op == 2)  {
        if (!chan)  {
            if (kei2001_write_data (instrID, ":ROUT:SCAN:LSEL NONE\r\n", 22) != 0)
                return kei2001_err;
        }
        else  {
            Fmt (cmd, "%s<:ROUT:SCAN (@%d[b4])\r\n", chan);
            if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
                return kei2001_err;
        }
    }
    else  {
        Fmt (cmd, "%s<:res:aver:adv 0\r\n");
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
        Fmt (cmd, "%s<:volt:dc:aver:adv 0\r\n");
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
        if (!op)  {
            Fmt (cmd, "%s<:rout:scan:lsel rat\r\n");
            if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
                return kei2001_err;
        }
        else  {
            Fmt (cmd, "%s<:rout:scan:lsel delt\r\n");
            if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
                return kei2001_err;
        }
    }
    if (kei2001_write_data (instrID, ":init:cont?\r\n", 13) != 0)
        return kei2001_err;
    if (kei2001_read_data (instrID, cmd, MAX_CMD_LENGTH) != 0)
        return kei2001_err;
    if (Scan (cmd, "%s>%i[b4]", &cont) != 1)  {
        kei2001_err = 236;
        return kei2001_err;
    }
    if (cont == 1)
        if (kei2001_write_data (instrID, ":init:cont 0\r\n", 14) != 0)
            return kei2001_err;
    if (kei2001_write_data (instrID, ":read?\r\n", 8) != 0)
        return kei2001_err;
    if (kei2001_read_data (instrID, cmd, MAX_CMD_LENGTH) != 0)
        return kei2001_err;
    if (cont == 1)
        if (kei2001_write_data (instrID, ":init:cont 1\r\n", 14) != 0)
            return kei2001_err;
    if (Scan (cmd, "%s>%f", result) != 1)  {
        kei2001_err = 236;
        return kei2001_err;
    }
    /*  Check the Event Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*  ======================================================================== */
/*   This function initiates a triggering cycle                              */
/* ========================================================================= */
int PUBLIC kei2001_start_meas (int instrID,int op,int chan)
{
/*  int cont;
    int adv_filter_dc;
    int adv_filter_res;
*/
    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (op, 0, 2,  -2) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (chan,  -1, 10,  -3) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    if (op == 2)  {
        if (!chan)  {
            if (kei2001_write_data (instrID, ":ROUT:SCAN:LSEL NONE\r\n", 22) != 0)
                return kei2001_err;
        }
        else  if (chan ==  -1)  {
            Fmt (cmd, "%s<:ROUT:SCAN:LSEL INT\r\n");
            if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
                return kei2001_err;
        }
        else  {
            Fmt (cmd, "%s<:ROUT:SCAN (@%d[b4])\r\n", chan);
            if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
                return kei2001_err;
        }
    }
    else  {
        Fmt (cmd, "%s<:volt:dc:aver:adv 0\r\n");
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
        Fmt (cmd, "%s<:res:aver:adv 0\r\n");
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
        if (!op)  {
            Fmt (cmd, "%s<:rout:scan:lsel rat\r\n");
            if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
                return kei2001_err;
        }
        else  {
            Fmt (cmd, "%s<:rout:scan:lsel delt\r\n");
            if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
                return kei2001_err;
        }
    }
    if (kei2001_write_data (instrID, ":init\r\n", 7) != 0)
        return kei2001_err;
    /*  Check the Event Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*  ======================================================================== */
/*   This function checks whether an operation has been completed             */
/* ========================================================================= */
int PUBLIC kei2001_check_meas (int instrID,int *result)
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    if (kei2001_write_data (instrID, "*OPC\r\n", 6) != 0)
        return kei2001_err;
    *result = 0;
    if (kei2001_write_data (instrID, "*ESR?\r\n", 7) != 0)
        return kei2001_err;
    if (kei2001_read_data (instrID, cmd, MAX_CMD_LENGTH) != 0)
        return kei2001_err;
    Scan (cmd, "%s>%d[b4]", result);
    /*  Check the Event Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*  ======================================================================== */
/*   This function retrieves the readings from buffer                        */
/* ========================================================================= */
int PUBLIC kei2001_retrieve_meas (int instrID,int *size,double result[])
{
    int file_handle;
    int err;

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:trac:poin?\r\n");
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    if (kei2001_read_data (instrID, cmd, MAX_CMD_LENGTH) != 0)
        return kei2001_err;
    if (Scan (cmd, "%s>%d[b4]", size) != 1)  {
        kei2001_err = 236;
        return kei2001_err;
    }
    /*  transfer the buffer data into a file  */
    Fmt (cmd, "%s<:form asc\r\n");
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:form:elem read\r\n");
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    if (kei2001_write_data (instrID, ":trac:data?\r\n", 13) != 0)
        return kei2001_err;
    if (kei2001_read_data_file (instrID, "temp.rd") != 0)
        return kei2001_err;
    file_handle = OpenFile ("temp.rd", 1, 2, 1);
    err = ScanFile (file_handle, "%s>%*f[x]", *size, result);
    if (err != 1)  {
        kei2001_err = 100;
        CloseFile (file_handle);
        file_handle = OpenFile ("temp.rd", 2, 0, 1);
        CloseFile (file_handle);
        return kei2001_err;
    }
    CloseFile (file_handle);
    file_handle = OpenFile ("temp.rd", 2, 0, 1);
    /* truncate the file to 0 byte */
    CloseFile (file_handle);
    /*  Check the Event Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*  ======================================================================== */
/*   This function transfers the buffer to a file                            */
/* ========================================================================= */
int PUBLIC kei2001_to_file (int instrID,char file_name[],int md,int *size)
{
    int file_handle;
    char temp[100];

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    if (md != 2 && md != 0)  {
        kei2001_err = 30;
        return kei2001_err;
    }
    Fmt (temp, "%s<%s.log", file_name);
    if (md == 2)  {
        file_handle = OpenFile (temp, 1, 2, 1);
        if (file_handle >= 0)  {
            /*  file exists */
            CloseFile (file_handle);
            return 100;
        }
    }
    file_handle = OpenFile (temp, 2, 0, 1);
    if (file_handle < 0)  {
        CloseFile (file_handle);
        return 101;
    }
    Fmt (cmd, "%s<:trac:poin?\r\n");
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    if (kei2001_read_data (instrID, cmd, MAX_CMD_LENGTH) != 0)
        return kei2001_err;
    Scan (cmd, "%s>%d[b4]", size);
    FmtFile (file_handle, "%s<%d[b4],", *size);
    /*  transfer the buffer data into a file  */
    Fmt (cmd, "%s<:form asc\r\n");
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:form:elem read,rnum,time,chan,stat,unit\r\n");
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    if (kei2001_write_data (instrID, ":trac:data?\r\n", 13) != 0)
        return kei2001_err;
    if (kei2001_read_data (instrID, cmd, MAX_CMD_LENGTH) != 0)
        return kei2001_err;
    while (ibcnt == (long)MAX_CMD_LENGTH && cmd[MAX_CMD_LENGTH - 1] != 10)  {
        WriteFile (file_handle, cmd, MAX_CMD_LENGTH);
        if (kei2001_read_data (instrID, cmd, MAX_CMD_LENGTH) != 0)
            return kei2001_err;
    }
    WriteFile (file_handle, cmd, MAX_CMD_LENGTH);
    CloseFile (file_handle);
    /*  Check the Event Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*  ======================================================================== */
/*   This function retrieves readings from a file                            */
/* ========================================================================= */
int PUBLIC kei2001_ret_read (char file_name[],int offset,int skip,int size, 
                            double reading[],int unit[],int stat[], 
                            double tm[],int chan[])
{
    int file_handle;
    int err;
    int i;
    int j;
    int point_ct;
    char temp[100];
    char st[2];

    if (kei2001_invalid_integer_range (offset, 0, 29999,  -2) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (skip, 0, 29999,  -3) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (size, 0, 30000,  -4) != 0)
        return kei2001_err;
    Fmt (temp, "%s<%s.log", file_name);
    file_handle = OpenFile (temp, 1, 2, 1);
    if (file_handle < 0)  {
        kei2001_err = 101;
        return kei2001_err;
    }
    ScanFile (file_handle, "%s>%d[b4x]", &point_ct);
    if (offset + (size - 1) * (skip + 1) + 1 > point_ct)  {
        kei2001_err = 102;
        return kei2001_err;
    }
    for (i = 0; i < offset; i++)
        err = ScanFile (file_handle, "%s>%s[t44dx]%s[t44dx]%s[t44dx]%s[t44dx]");
    for (i = 0; i < size - 1; i++)  {
        err = ScanFile (file_handle, "%s>%f%s[w1]%s[t44x]%f%s[t44dx]%s[t44dx]%d[b4]%s[t44dx]", &reading[i], st, temp, &tm[i]
        , &chan[i]);
        if (!CompareStrings (st, 0, "N", 0, 0))
            stat[i] = 0;
        else  if (!CompareStrings (st, 0, "O", 0, 0))
            stat[i] = 1;
        else  if (!CompareStrings (st, 0, "R", 0, 0))
            stat[i] = 2;
        if (!CompareStrings (temp, 0, "VDC", 0, 0))
            unit[i] = 0;
        else  if (!CompareStrings (temp, 0, "VAC", 0, 0))
            unit[i] = 1;
        else  if (!CompareStrings (temp, 0, "ADC", 0, 0))
            unit[i] = 2;
        else  if (!CompareStrings (temp, 0, "AAC", 0, 0))
            unit[i] = 3;
        else  if (!CompareStrings (temp, 0, "OHM", 0, 0))
            unit[i] = 4;
        else  if (!CompareStrings (temp, 0, "OHM4W", 0, 0))
            unit[i] = 5;
        else  if (!CompareStrings (temp, 0, "HZ", 0, 0))
            unit[i] = 6;
        else  if (!CompareStrings (temp, 0, "C", 0, 0))
            unit[i] = 7;
        else  if (!CompareStrings (temp, 0, "F", 0, 0))
            unit[i] = 8;
        else  if (!CompareStrings (temp, 0, "K", 0, 0))
            unit[i] = 9;
        for (j = 0; j < skip; j++)
            ScanFile (file_handle, "%s>%s[t44dx]%s[t44dx]%s[t44dx]%s[t44dx]");
    }
    ScanFile (file_handle, "%s>%f%s[w1]%s[t44x]%f%s[t44dx]%s[t44dx]%d[b4]%s[t44dx]", &reading[i], st, temp, &tm[i], &chan[i]
    );
    if (!CompareStrings (st, 0, "N", 0, 0))
        stat[i] = 0;
    else  if (!CompareStrings (st, 0, "O", 0, 0))
        stat[i] = 1;
    else  if (!CompareStrings (st, 0, "R", 0, 0))
        stat[i] = 2;
    if (!CompareStrings (temp, 0, "VDC", 0, 0))
        unit[i] = 0;
    else  if (!CompareStrings (temp, 0, "VAC", 0, 0))
        unit[i] = 1;
    else  if (!CompareStrings (temp, 0, "ADC", 0, 0))
        unit[i] = 2;
    else  if (!CompareStrings (temp, 0, "AAC", 0, 0))
        unit[i] = 3;
    else  if (!CompareStrings (temp, 0, "OHM", 0, 0))
        unit[i] = 4;
    else  if (!CompareStrings (temp, 0, "OHM4W", 0, 0))
        unit[i] = 5;
    else  if (!CompareStrings (temp, 0, "HZ", 0, 0))
        unit[i] = 6;
    else  if (!CompareStrings (temp, 0, "C", 0, 0))
        unit[i] = 7;
    else  if (!CompareStrings (temp, 0, "F", 0, 0))
        unit[i] = 8;
    else  if (!CompareStrings (temp, 0, "K", 0, 0))
        unit[i] = 9;
    CloseFile (file_handle);
    return 0;
}

/* ========================================================================== */
/*  Function: Buffer Reading Calculation                                      */
/*  Description: Specify operations on readings stored in the buffer and      */
/*               return the result.                                           */
/* ========================================================================== */
int PUBLIC kei2001_calc_buf (int instrID,int state,int frmt,double *result)
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (state, 0, 1,  -2) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (frmt, 0, 5,  -3) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:calc2:stat %d[b4]\r\n", state);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:calc2:form %s\r\n", calc2_format_str[frmt]);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:calc2:imm?\r\n");
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    if (kei2001_read_data (instrID, cmd, MAX_CMD_LENGTH) != 0)
        return kei2001_err;
    Scan (cmd, "%s>%f", result);
    /*  Check the Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*  ======================================================================== */
/*   This function resets the triggering cycle                               */
/* ========================================================================= */
int PUBLIC kei2001_trig_reset (int instrID)
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:abor\r\n");
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    /*  Check the Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*  ======================================================================== */
/*   This function resets the instrument to the Power-On reset state         */
/* ========================================================================= */
int PUBLIC kei2001_preset (int instrID)
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    if (kei2001_write_data (instrID, ":syst:pres\r\n", 12) != 0)
        return kei2001_err;
    /*  Check the Event Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*  ======================================================================== */
/*   This function resets the instrument to the *RST default state           */
/* ========================================================================= */
int PUBLIC kei2001_reset (int instrID)
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    if (kei2001_write_data (instrID, "*RST\r\n", 6) != 0)
        return kei2001_err;
    /*  Check the Event Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*  ======================================================================== */
/*   This function waits until the previous commands are executed            */
/* ========================================================================= */
int PUBLIC kei2001_wait (int instrID)
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    if (kei2001_write_data (instrID, "*WAI\r\n", 6) != 0)
        return kei2001_err;
    /*  Check the Event Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*  ======================================================================== */
/*   This function issues a bus trigger                                      */
/* ========================================================================= */
int PUBLIC kei2001_bus_trig (int instrID)
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    if (kei2001_write_data (instrID, "*TRG\r\n", 6) != 0)
        return kei2001_err;
    /*  Check the Event Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*  ======================================================================== */
/*   This function opens a specific channel                                  */
/* ========================================================================= */
int PUBLIC kei2001_open_chan (int instrID,int chan)
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (chan, 1, 10,  -2) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:rout:open (@%d[b4])\r\n", chan);
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    /*  Check the Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*   ======================================================================  */
/*    This function sets the polarity of the digital output lines            */
/*   ======================================================================  */
int PUBLIC kei2001_dig_polarity (int instrID,int partern)
{
    int level[4];
    int i;

    level[0] = partern & 1;
    level[1] = (partern & 2) >> 1;
    level[2] = (partern & 4) >> 2;
    level[3] = (partern & 8) >> 3;
    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (partern, 0, 15,  -2) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    for (i = 0; i < 4; i++)  {
        Fmt (cmd, ":outp:ttl%d[b4]:lsen %s", i + 1, polarity_str[level[i]]);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    /*  Check the Event Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*   ======================================================================   */
/*    This function sets the TTL level of the digital output lines            */
/*   ======================================================================   */
int PUBLIC kei2001_dig_out (int instrID,int partern)
{
    int level[4];
    int i;

    level[0] = partern & 1;
    level[1] = (partern & 2) >> 1;
    level[2] = (partern & 4) >> 2;
    level[3] = (partern & 8) >> 3;
    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (partern, 0, 15,  -2) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    for (i = 0; i < 4; i++)  {
        Fmt (cmd, "%s<%s%d[b4] %d[b4]", ":sour:ttl", i + 1, level[i]);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    /*  Check the Event Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*   ======================================================================   */
/*    This function reads the TTL level of the single digital lines           */
/*   ======================================================================   */
int PUBLIC kei2001_dig_in (int instrID,int *reading)
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<%s", ":sens2:ttl:data?");
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    kei2001_err = 0;
    if (kei2001_read_data (instrID, cmd, 3) != 0)
        return kei2001_err;
    Scan (cmd, "%s>%d[b4]", reading);
    /*  Check the Event Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*   =========================================================================   */
/*    This function displays a user defined message on the instrument display    */
/*   =========================================================================   */
int PUBLIC kei2001_disp_msg (int instrID,int fun,char message[])
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (fun, 0, 2,  -2) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (StringLength (message), 0, 20,  -3) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    FillBytes (cmd, 0, 100, 0);
    switch (fun)  {
    case 0:
        Fmt (cmd, "%s<%s", ":disp:text:stat off");
        break;
    case 1:
        Fmt (cmd, "%s<%s", ":disp:text:stat on");
        break;
    case 2:
        Fmt (cmd, "%s<%s", ":disp:wind2:text:stat on");
        break;
    }
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    switch (fun)  {
    case 0:
        Fmt (cmd, "%s<%s", ":disp:wind2:text:stat off");
        break;
    case 1:
        Fmt (cmd, "%s<%s%s\'", ":disp:text:data \'", message);
        break;
    case 2:
        Fmt (cmd, "%s<%s%s\'", ":disp:wind2:text:data \'", message);
        break;
    }
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    /*  Check the Event Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*  ======================================================================== */
/*   This function queries the instrument error queqe                        */
/* ========================================================================= */
int PUBLIC kei2001_read_instr_err (int instrID,int *err_code,char err_msg[])
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    if (kei2001_write_data (instrID, ":SYST:ERR?\r\n", 12) != 0)
        return kei2001_err;
    if (kei2001_read_data (instrID, cmd, MAX_CMD_LENGTH) != 0)
        return kei2001_err;
    if (Scan (cmd, "%s>%i[b4]%s[t44xd]%s[t34xd]%s[t34x]", err_code, err_msg) != 4)  {
        kei2001_err = 236;
        return kei2001_err;
    }
    /*  Check the Event Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*  ======================================================================== */
/*   This function clears the readings in the buffer                         */
/* ========================================================================= */
int PUBLIC kei2001_buf_clear (int instrID)
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    Fmt (cmd, "%s<:trac:clear\r\n");
    if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
        return kei2001_err;
    /*  Check the Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/* ========================================================================= */
/*  Function: Save/Recall Setup                                                         */
/*  Purpose:  This function saves or recalls instrument setup.               */
/* ========================================================================= */
int PUBLIC kei2001_sav_rec_setup (int instrID,int op,int loc)
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (op, 0, 1,  -2) != 0)
        return kei2001_err;
    if (kei2001_invalid_integer_range (loc, 0, sav_size[kei2001_memory_size[instrID]],  -3) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    if (!op)  {
        Fmt (cmd, "%s<*SAV %d[b4]\r\n", loc);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    else  {
        Fmt (cmd, "%s<*RCL %d[b4]\r\n", loc);
        if (kei2001_write_data (instrID, cmd, NumFmtdBytes ()) != 0)
            return kei2001_err;
    }
    /*  Check the Status Byte of the instrument for proper execution.  */
    if (kei2001_check_instr_status (instrID) != 0)
        return kei2001_err;
    return kei2001_err;
}

/*    ==========================================================================    */
/*     Function:     Manual Control                                                 */
/*     Description:  This function returns the instrument to local front panel      */
/*                   control.                                                       */
/*    ==========================================================================    */
int PUBLIC kei2001_goto_local (int instrID)
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    if (kei2001_loc_remote (instrID, 0) != 0)
        return kei2001_err;
    return kei2001_err;
}

/* ========================================================================= */
/*  Function: Close                                                          */
/*  Purpose:  This function closes the instrument.                           */
/* ========================================================================= */
int PUBLIC kei2001_close (int instrID)
{

    if (kei2001_invalid_integer_range (instrID, 1, kei2001_MAX_INSTR,  -1) != 0)
        return kei2001_err;
    if (kei2001_device_closed (instrID) != 0)
        return kei2001_err;
    kei2001_close_instr (instrID);
    return kei2001_err;
}

/* = UTILITY ROUTINES ====================================================== */
/* ========================================================================= */
/*  Function Name:     Open Instrument                                       */
/*  Description:       Locates and initializes an entry in the Instrument    */
/*                     Table and the GPIB device table for the instrument.   */
/*  Return Value:      Error (Instr. Driver Status)                          */
/* ========================================================================= */
int kei2001_open_instr (int addr,int *id)
{
    int i;
    int instrID;

    instrID = 0;
    kei2001_err = 0;
    /*  Check to see if the instrument is already in the Instrument Table.  */
    for (i = 1; i <= kei2001_MAX_INSTR; i++)
        if (address[i] == addr)  {
            instrID = i;
            i = kei2001_MAX_INSTR;
        }
    /*  If it is not in the instrument table, open an entry for the instrument  */
    if (instrID <= 0)
        for (i = 1; i <= kei2001_MAX_INSTR; i++)
            if (!address[i])  {
                instrID = i;
                address[instrID] = addr;
                instr_cnt++;
                i = kei2001_MAX_INSTR;
            }
    /*  If an entry could not be opened in the Instrument Table, return an error */
    if (instrID <= 0)  {
        kei2001_err = 220;
        return kei2001_err;
    }
    /*  Initialize the GPIB interface.  */
    if (bd[instrID] <= 0)  {
        if (instr_cnt <= 1)
            CloseInstrDevs ("kei2001");
        bd[instrID] = OpenDev ("", "kei2001");
        if (bd[instrID] <= 0)  {
            kei2001_err = 220;
            return kei2001_err;
        }
    }
    /*   Change the primary address of the device.  */
    if (ibpad (bd[instrID], addr) & 0x8000)  {
        kei2001_err = 233;
        return kei2001_err;
    }
    *id = instrID;
    return kei2001_err;
}

/* ========================================================================= */
/*  Function Name:     Close Instrument                                      */
/*  Description:       Closes the instrument by removing it from the GPIB    */
/*                     device table, and setting the GPIB address, bd, and   */
/*                     interface to zero in the Instrument Table.            */
/*  Return Value:      0 - successful        -1 - unsuccessful               */
/* ========================================================================= */
int kei2001_close_instr (int instrID)
{

    if (bd[instrID] != 0)  {
        CloseDev (bd[instrID]);
        bd[instrID] = 0;
        address[instrID] = 0;
        instr_cnt--;
    }
    else  {
        kei2001_err = 221;
        return  -1;
    }
    return 0;
}

/* ========================================================================= */
/*  Function: Invalid real Range                                             */
/*  Purpose:  This function checks a double to see if it lies between a      */
/*            minimum and maximum value.  If the value is out of range, set  */
/*            the global error variable to the value err_code.  If the       */
/*            value is OK, error = 0.                                        */
/* ========================================================================= */
int kei2001_invalid_real_range (double val,double min,double max,int err_code)
{

    if (val < min || val > max)  {
        kei2001_err = err_code;
        return  -1;
    }
    return 0;
}

/* ========================================================================= */
/*  Function Name:     Invalid Integer Range                                 */
/*  Description:       Checks an integer to see if it lies between a         */
/*                     minimum and maximum value.  If the value is out of    */
/*                     range, sets the Error variable to the value           */
/*                     err_code.  If the value is OK, Error = 0.             */
/*  Return Value:      Error (Instr. Driver Status)                          */
/* ========================================================================= */
int kei2001_invalid_integer_range (int val,int min,int max,int err_code)
{

    if (val < min || val > max)
        kei2001_err = err_code;
    else
        kei2001_err = 0;
    return kei2001_err;
}

/* ========================================================================= */
/*  Function: Invalid Long Range                                             */
/*  Purpose:  This function checks a long int to see if it lies between a    */
/*            minimum and maximum value.  If the value is out of range, set  */
/*            the global error variable to the value err_code.  If the       */
/*            value is OK, error = 0.                                        */
/* ========================================================================= */
int kei2001_invalid_long_range (long val,long min,long max,int err_code)
{

    if (val < min || val > max)  {
        kei2001_err = err_code;
        return  -1;
    }
    return 0;
}

/* ========================================================================= */
/*  Function Name:     Device Closed?                                        */
/*  Description:       Checks to see if the instrument has been initialized. */
/*  Return Value:      Error (Instr. Driver Status)                          */
/* ========================================================================= */
int kei2001_device_closed (int instrID)
{

    if (bd[instrID] <= 0)
        kei2001_err = 232;
    else
        kei2001_err = 0;
    return kei2001_err;
}

/* ========================================================================= */
/*  Function Name:     Read Message                                          */
/*  Description:       Reads a buffer of data from the instrument.           */
/*  Return Value:      Error (Instr. Driver Status)                          */
/* ========================================================================= */
int kei2001_read_data (int instrID,char *buf,int cnt)
{

    if (ibrd (bd[instrID], buf, (long)cnt) & 0x8000)
        kei2001_err = 231;
    else
        kei2001_err = 0;
    return kei2001_err;
}

/* ========================================================================= */
/*  Function Name:     Write Message                                         */
/*  Description:       Writes a buffer of data to the instrument.            */
/*  Return Value:      Error (Instr. Driver Status)                          */
/* ========================================================================= */
int kei2001_write_data (int instrID,char *buf,int cnt)
{

    if (ibwrt (bd[instrID], buf, (long)cnt) & 0x8000)
        kei2001_err = 230;
    else
        kei2001_err = 0;
    return kei2001_err;
}

/*  ==========================================================================  */
/*   This function is called by the init routine to initialize global arrays    */
/*   This routine should be modified for each instrument to include             */
/*   instrument-dependent commmand arrays.                                      */
/*  ==========================================================================  */
void kei2001_setup_arrays (void)
{

    function[0] = "volt:dc";
    /*   DC Voltage measurement  */
    function[1] = "volt:ac";
    /*   AC Voltage measurement  */
    function[2] = "res";
    /*   2-wire Resistance measurement  */
    function[3] = "fres";
    /*   4-wire Resistance measurement  */
    function[4] = "curr:dc";
    /*   DC Current measurement  */
    function[5] = "curr:ac";
    /*   AC Current measurement  */
    function[6] = "freq";
    function[7] = "temp";
    aperture[0] = 1.0;
    aperture[1] = 0.01;
    aperture[2] = 0.1;
    aperture[3] = 10.0;
    scan_func[0] = "int";
    scan_func[1] = "rat";
    scan_func[2] = "delt";
    trig_cmd[0] = "hold";
    trig_cmd[1] = "imm";
    trig_cmd[2] = "man";
    trig_cmd[3] = "bus";
    trig_cmd[4] = "tlin";
    trig_cmd[5] = "ext";
    trig_cmd[6] = "tim";
    func_input_range[0] = 775.0;
    func_input_range[1] = 1100.0;
    func_input_range[2] = 1050000000.0;
    func_input_range[3] = 210000.0;
    func_input_range[4] = 2.1;
    func_input_range[5] = 2.1;
    dc_str[0] = "DC";
    dc_str[1] = "AC";
    freq_source_str[0] = "VOLT";
    freq_source_str[1] = "CURR";
    freq_def_range[1] = 0.001;
    freq_def_range[0] = 10.0;
    freq_source_range[1] = 1.0;
    freq_source_range[0] = 1000.0;
    tc_type_str = "JTKERSB";
    rtd_type_str[0] = "PT385";
    rtd_type_str[1] = "PT3916";
    trans_type_str[0] = "RTD";
    trans_type_str[1] = "FRTD";
    trans_type_str[2] = "TC";
    temp_unit_str = "CFK";
    ac_unit_str[0] = "V";
    ac_unit_str[1] = "DB";
    ac_unit_str[2] = "DBM";
    dbm_imp[0] = 50.0;
    dbm_imp[1] = 75.0;
    dbm_imp[2] = 93.0;
    dbm_imp[3] = 132.0;
    dbm_imp[4] = 300.0;
    dbm_imp[5] = 600.0;
    auto_zero_mode_str[0] = "NORM";
    auto_zero_mode_str[1] = "SYNC";
    acq_method_str[0] = "NORM";
    acq_method_str[1] = "BURS";
    filter_type_str[0] = "rep";
    filter_type_str[1] = "mov";
    auto_filter_str[0] = "0";
    auto_filter_str[1] = "1";
    auto_filter_str[2] = "once";
    source_str[0] = "hold";
    source_str[1] = "imm";
    source_str[2] = "man";
    source_str[3] = "bus";
    source_str[4] = "tlin";
    source_str[5] = "ext";
    source_str[6] = "tim";
    buf_size_range[0] = 250L;
    buf_size_range[1] = 1400L;
    buf_size_range[2] = 6000L;
    buf_size_range[3] = 850L;
    buf_size_range[4] = 7000L;
    buf_size_range[5] = 30000L;
    egroup_str[0] = "full";
    egroup_str[1] = "comp";
    buf_src_str[0] = "sens";
    buf_src_str[1] = "calc";
    buf_src_str[2] = "none";
    buf_cntrl_mode_str[0] = "nev";
    buf_cntrl_mode_str[1] = "next";
    buf_cntrl_mode_str[2] = "alw";
    buf_cntrl_mode_str[3] = "pret";
    pretrig_src_str[0] = "ext";
    pretrig_src_str[1] = "tlin";
    pretrig_src_str[2] = "bus";
    pretrig_src_str[3] = "man";
    math_format_str[0] = "mxb";
    math_format_str[1] = "perc";
    math_format_str[2] = "none";
    calc2_format_str[0] = "none";
    calc2_format_str[1] = "mean";
    calc2_format_str[2] = "sdev";
    calc2_format_str[3] = "max";
    calc2_format_str[4] = "min";
    calc2_format_str[5] = "pkpk";
    chan_fun_str[0] = "\'volt:dc\'";
    chan_fun_str[1] = "\'volt:ac\'";
    chan_fun_str[2] = "\'temp\'";
    chan_fun_str[3] = "\'freq\'";
    chan_fun_str[4] = "\'res\'";
    chan_fun_str[5] = "\'fres\'";
    chan_fun_str[6] = "\'none\'";
    source_bypass_str[0] = "sour";
    source_bypass_str[1] = "acc";
    rat_del_fun_str[0] = "volt:dc";
    rat_del_fun_str[1] = "res";
    rat_del_fun_str[2] = "fres";
    polarity_str[0] = "alow";
    polarity_str[1] = "ahig";
    sav_size[0] = 0;
    sav_size[1] = 4;
    sav_size[2] = 9;
}

/* ========================================================================= */
/*  Function Name:     Check Instrument Status                               */
/*  Description:       Checks the status of the Instrument                   */
/*                     and sets kei2001 Error accordingly.                  */
/*  Return Value:      kei2001 Error (Instrument Driver Status)             */
/* ========================================================================= */
int kei2001_check_instr_status (int instrID)
{
    int event_status;

    event_status = 0;
    kei2001_err = 0;
    /*  Query the instrument for the Event Status Register contents.  */
    if (ibwrt (bd[instrID], "*ESR?\r\n", 7L) & 0x8000)  {
        kei2001_err = 230;
        return kei2001_err;
    }
    if (ibrd (bd[instrID], cmd, (long)MAX_CMD_LENGTH) & 0x8000)  {
        kei2001_err = 231;
        return kei2001_err;
    }
    if (Scan (cmd, "%s>%d[b4]", &event_status) != 1)  {
        kei2001_err = 236;
        return kei2001_err;
    }
    if ((event_status & 0x3c) != 0)
        kei2001_err = 300;
    return kei2001_err;
}

/* ========================================================================= */
/*  Function: Read Data File                                                 */
/*  Purpose:  This function reads a buffer of data from the instrument and   */
/*            stores it to the file specified by "filename".  Filename must  */
/*            either be a string, such as "C:\lw\instr\file" or a pointer    */
/*            to such a string.  The return value is equal to the global     */
/*            error variable.                                                */
/* ========================================================================= */
int kei2001_read_data_file (int instrID,char *filename)
{

    if (ibrdf (bd[instrID], filename) & 0x8000)
        kei2001_err = 229;
    else
        kei2001_err = 0;
    return kei2001_err;
}

/* ========================================================================= */
/*  Function Name: Local/Remote                                              */
/*  Description:   Puts the instrument into either local or remote mode.     */
/*  Return Value:  Error (Instr. Driver Status)                              */
/* ========================================================================= */
int kei2001_loc_remote (int instrID,int loc_remote)
{

    kei2001_err = 0;
    switch (loc_remote)  {
    case 0:
        if (ibloc (bd[instrID]) & 0x8000)
            kei2001_err = 234;
        break;
    case 1:
        if (ibwrt (bd[instrID], "", 0L) & 0x8000)
            kei2001_err = 235;
        break;
    default:
        kei2001_err = 234;
        break;
    }
    return kei2001_err;
}

/*****************************************************************************/
/*=== END INSTRUMENT DRIVER SOURCE CODE =====================================*/
/*****************************************************************************/
